@@ -12,6 +12,9 @@ import jinjasql
 import xmltodict
 import pyparsing as pp
 import subprocess
+import openpyxl
+from openpyxl.utils.dataframe import dataframe_to_rows
+
 from IPython.core.magic import Magics, magics_class, line_cell_magic
 from IPython.display import display, HTML
 from IPython import get_ipython
@@ -92,6 +95,11 @@ def identity(value):
     return Markup(f""" "{value.replace('"', '""')}" """)
 
 
+def fields(values):
+    fields = ",".join(f""" "{value.replace('"', '""')}" """ for value in values)
+    return Markup(f"{fields}")
+
+
 class Session:
     def __init__(
         self,
@@ -102,6 +110,7 @@ class Session:
         df_viewer=None,
         df_viewer_kw=None,
         datasette_url=None,
+        cell_magic_output=False
     ):
         self.schema = schema
         self.dburi = dburi
@@ -109,6 +118,7 @@ class Session:
         self.df_viewer_kw = df_viewer_kw or {}
 
         self.datasette_url = datasette_url
+        self.cell_magic_output = cell_magic_output
 
         if datasette_url:
             self.database_type = "datasette"
@@ -127,7 +137,10 @@ class Session:
 
         self.jinjarender = jinjasql.JinjaSql(param_style=param_style)
         self.jinjarender.env.filters["s"] = safe
+        self.jinjarender.env.filters["sql"] = safe
         self.jinjarender.env.filters["i"] = identity
+        self.jinjarender.env.filters["ident"] = identity
+        self.jinjarender.env.filters["fields"] = fields
 
         self.overwrite = overwrite
         if schema:
@@ -186,20 +199,26 @@ class Session:
             return results
 
     def create_table(self, table, sql, params=None):
+        return self.create_relation(table, sql, 'TABLE', params)
+
+    def create_view(self, view, sql, params=None):
+        return self.create_relation(view, sql, 'VIEW', params)
+
+    def create_relation(self, relation, sql, type, params=None):
         self.run_sql(
             """
-            DROP TABLE IF EXISTS {table};
+            DROP {type} IF EXISTS {relation};
         """.format(
-                table=table
+                relation=relation, type=type
             )
         )
 
         sql = """
-            CREATE TABLE {table}
+            CREATE {type} {relation}
             AS
             {sql}
         """.format(
-            table=table, sql=sql
+            relation=relation, sql=sql, type=type
         )
         return self.run_sql(sql, params=params)
 
@@ -508,15 +527,27 @@ class Noteql(Magics):
         create = pp.Keyword("create", caseless=True).suppress() + (
             pp.Word(pp.alphanums + "_") | pp.QuotedString('"', escQuote='"', unquoteResults=False)
         )("create")
+        view = pp.Keyword("view", caseless=True).suppress() + (
+            pp.Word(pp.alphanums + "_") | pp.QuotedString('"', escQuote='"', unquoteResults=False)
+        )("view")
         df_arrows = (pp.Word(pp.alphanums + "_") + pp.Keyword("<<").suppress())("df")
 
         session = (
             pp.Keyword("session", caseless=True).suppress()
-            + pp.Optional(pp.Keyword("as", caseless=True).suppress())
             + pp.Word(pp.alphanums + "_")("session")
         )
 
-        commands = [arg_params, create, df_arrows, session]
+        csv = (
+            pp.Keyword("csv", caseless=True).suppress()
+            + (pp.QuotedString("'", escQuote="'") | pp.Word(pp.printables))
+        )('csv')
+
+        excel = (
+            pp.Keyword("excel", caseless=True).suppress()
+            + (pp.QuotedString("'", escQuote="'") | pp.Word(pp.printables))
+        )('excel')
+
+        commands = [arg_params, create, view, df_arrows, session, csv, excel]
 
         for cmd_string in [
             "df",
@@ -536,7 +567,7 @@ class Noteql(Magics):
                 + pp.Keyword(cmd_string, caseless=True).suppress()
             )
 
-        title = pp.Keyword("title", caseless=True).suppress() + pp.QuotedString("'", escQuote="'")("title")
+        title = pp.Keyword("title", caseless=True).suppress() + (pp.QuotedString("'", escQuote="'") | pp.Word(pp.printables))("title")
         nojinja = (pp.Keyword("nojinja", caseless=True) | pp.Keyword("noj", caseless=True))("nojinja")
         show = pp.Keyword("show", caseless=True)("show")
         rest = pp.Word(pp.printables)("rest")
@@ -578,6 +609,7 @@ class Noteql(Magics):
         actions = {}
         arg_params = {}
         jinja = True
+        title = None
 
         for item in parsed_line:
 
@@ -603,8 +635,11 @@ class Noteql(Magics):
 
             for command in [
                 "df",
+                "csv",
+                "excel",
                 "sql",
                 "create",
+                "view",
                 "row",
                 "rows",
                 "col",
@@ -627,7 +662,8 @@ class Noteql(Magics):
                 jinja = False
 
             if item.getName() == "title":
-                session.show_title(item[0])
+                title = item[0]
+                session.show_title(title)
 
         params = None
         if jinja:
@@ -650,6 +686,8 @@ class Noteql(Magics):
         df = None
 
         df_name = actions.get("df")
+        csv_file = actions.get("csv")
+        excel_file = actions.get("excel")
         col_name = actions.get("col")
         cols_name = actions.get("cols")
         row_name = actions.get("row")
@@ -664,6 +702,8 @@ class Noteql(Magics):
         if any(
             [
                 show,
+                csv_file,
+                excel_file,
                 df_name,
                 col_name,
                 cols_name,
@@ -703,6 +743,29 @@ class Noteql(Magics):
             if headings_name:
                 ns[headings_name] = list(df.columns)
 
+            if csv_file:
+                df.to_csv(csv_file, index=False)
+
+            if excel_file:
+                if title:
+                    ws_title = title
+                else:
+                    ws_title = 'Sheet'
+
+                if os.path.exists(excel_file):
+                    wb = openpyxl.load_workbook(filename=excel_file)
+                    if ws_title in wb.sheetnames:
+                        raise Exception(f"%%nql error, sheet name {ws_title} already exists. Use a new sheet name by adding `TITLE 'My Sheet Name'`.")
+                    ws = wb.create_sheet(ws_title)
+                else:
+                    wb = openpyxl.Workbook()
+                    ws = wb.active
+                    ws.title = ws_title
+
+                for r in dataframe_to_rows(df, header=True, index=False):
+                    ws.append(r)
+                wb.save(excel_file)
+
             if show:
                 if session.df_viewer:
                     df = session.df_viewer(df, **session.df_viewer_kw)
@@ -713,11 +776,17 @@ class Noteql(Magics):
         if create_name:
             session.create_table(create_name, sql, params)
 
+        view_name = actions.get("view")
+
+        if view_name:
+            session.create_view(view_name, sql, params)
+
         return df
 
     @line_cell_magic
     def nql(self, line, cell=None):
 
+        session = self.find_session()
         if cell:
             dfs = []
             magic_line_parser, cell_parser = self.get_parsers()
@@ -731,12 +800,12 @@ class Noteql(Magics):
                 if not sql.strip():
                     print("Error in %%nql, empty sql statement")
                 dfs.append(self.execute_part(parsed_line[1:], sql))
-            return dfs
+            if session.cell_magic_output:
+                return dfs
         else:
             ns = self.shell.user_ns
             namespace_copy = ns.copy()
             params = None
-            session = self.find_session()
             line = line.replace("{", "{{").replace("}", "}}")
             sql, params = session.jinjarender.prepare_query(line, namespace_copy)
             if not params:
